open import Prelude

module DistrArray (placeCount : Nat) where  

  Data : Set
  Data = Nat

  Array : Nat -> Set
  Array n = Fin n -> Data

  Shape : Set
  Shape = List Nat

  data Heap : Shape -> Set where
    Empty : Heap Nil
    Alloc : {n : Nat} -> {ns : Shape} -> 
      Array n -> Heap ns -> Heap (Cons n ns)

  data Loc  : Nat -> Shape -> Set where
    Top  : {n : Nat} -> {ns : Shape} -> Loc n (Cons n ns)
    Pop  : forall {n k ns} -> Loc n ns -> Loc n (Cons k ns)

  Place : Set
  Place = Fin placeCount

  Region : Set
  Region = Nat

  Point : Region -> Set
  Point n = Fin n

  postulate
    distr : forall {n ns} -> Loc n ns -> Point n -> Place

  data DIO (a : Set) : Shape -> Place -> Shape -> Set where
    DReturn : {p : Place} -> {ns : Shape} -> a -> DIO a ns p ns
    DRead : forall {n ns ms} ->
      (l : Loc n ns) -> (pt : Point n) -> 
      (Data -> DIO a ns (distr l pt) ms) ->
      DIO a ns (distr l pt) ms
    DWrite : forall {n ns ms} ->
      (l : Loc n ns) -> (pt : Point n) -> Data -> 
      DIO a ns (distr l pt) ms ->
      DIO a ns (distr l pt) ms
    DNew : forall {p ns ms} ->
      (n : Nat) -> 
      (Loc n (Cons n ns) -> DIO a (Cons n ns) p ms) ->
      DIO a ns p ms
    At : forall {p ns ms ps} -> 
      (q : Place) -> DIO Unit ns q ms -> DIO a ms p ps -> DIO a ns p ps

  dreturn : {ns : Shape} {A : Set} {p : Place} -> A -> DIO A ns p ns
  dreturn a = DReturn a

  _>>>=_ : {ns ms ks : Shape} {A B : Set} {p : Place} ->
    DIO A ns p ms -> (A -> DIO B ms p ks) -> DIO B ns p ks
  (DReturn x) >>>= f = f x
  (DRead l i io) >>>= f = DRead l i (\x -> io x >>>= f)
  (DWrite l i x io) >>>= f = DWrite l i x (io >>>= f)
  (DNew n io) >>>= f = DNew n (\l -> io l >>>= f)
  At q there here >>>= f = At q there (here >>>= f)

  infixr 10 _>>>_

  _>>>_ : {ns ms ks : Shape} {A : Set} {p : Place} ->
    DIO Unit ns p ms -> DIO A ms p ks -> DIO A ns p ks
  x >>> f = x >>>= \y -> f

  at : forall {q ns ms} -> (p : Place) -> DIO Unit ns p ms -> DIO Unit ns q ms
  at p io = At p io (DReturn void)

  dreadArray : forall {ns n} -> (l : Loc n ns) 
    -> (i : Fin n) -> DIO Nat ns (distr l i) ns
  dreadArray l i = DRead l i DReturn

  dwriteArray : forall {ns n} -> (l : Loc n ns) -> (i : Fin n) -> Nat
    -> DIO Unit ns (distr l i) ns
  dwriteArray l i x = DWrite l i x (DReturn void)

  lookup : forall {n ns} -> Loc n ns -> Fin n -> Heap ns -> Data
  lookup  Top      i  (Alloc a _)  = a i
  lookup  (Pop k)  i  (Alloc _ h)  = lookup k i h

  updateArray : {n : Nat} -> Fin n -> Data -> Array n -> Array n
  updateArray i d a = \j -> if i == j then d else a j

  updateHeap : forall {n ns} -> 
    Loc n ns -> Fin n -> Data -> Heap ns -> Heap ns
  updateHeap Top      i x (Alloc a h)  = Alloc (updateArray i x a) h
  updateHeap (Pop k)  i x (Alloc a h)  = Alloc a (updateHeap k i x h)

  drun : forall {a ns ms} -> 
     (p : Place) -> DIO a ns p ms -> Heap ns -> Pair a (Heap ms)
  drun p (DReturn x) h                      =  pair x h
  drun .(distr l i) (DRead l i rd) h        =  drun (distr l i) (rd (lookup l i h)) h
  drun .(distr l i) (DWrite l i x wr) h     =  let h' = updateHeap l i x h
                                               in drun (distr l i) wr h'
  drun p (DNew n io) h                      =  drun p (io Top) (Alloc (\i -> Zero) h)
  drun p (At q io1 io2) h                   =  drun p io2 (snd (drun q io1 h))

  for : forall {n ns p} -> (Point n -> DIO Unit ns p ns) -> DIO Unit ns p ns
  for  {Succ k}   dio  = dio Fz >>> (for  {k}  (dio o Fs))
  for  {Zero}     dio  = dreturn void

  dmap : forall {n ns p} -> (Data -> Data) -> Loc n ns -> DIO Unit ns p ns
  dmap f l = for  (\i ->  at  (distr l i) (  dreadArray l i >>>= \x ->
                                             dwriteArray l i (f x)))
  forallplaces : forall {p ns} -> 
    ((q : Place) -> DIO Unit ns q ns) -> DIO Unit ns p ns
  forallplaces io = for (\i -> at i (io i))

  ateach : forall {n ns p} -> 
    (l : Loc n ns) -> ((pt : Point n) -> DIO Unit ns (distr l pt) ns) ->
    DIO Unit ns p ns
  ateach l io = for (\i -> at (distr l i) (io i))

  increment : forall {n ns p} -> 
    (l : Loc n ns) -> (i : Fin n) -> Nat -> (distr l i === p) -> DIO Unit ns p ns
  increment l i x Refl = dreadArray l i >>>= \y -> dwriteArray l i (x + y)

  sum : forall {n ns p} -> Loc n ns -> Loc 1 ns -> DIO Unit ns p ns
  sum l out = ateach l (\i ->  dreadArray l i >>>= \n ->
                               at (distr out Fz) (increment out Fz n Refl))

  psum : forall {n ns} ->
    (l : Loc n ns) -> (localSums : Loc placeCount ns) -> 
    ((i : Place) -> distr localSums i === i) -> 
    (out : Loc 1 ns) -> DIO Nat ns (distr out Fz) ns
  psum l localSums locDistr out =  
      ateach l (\ i -> (  dreadArray l i >>>= \n -> 
                          increment localSums (distr l i) n (locDistr (distr l i)))) 
      >>> sum localSums out
      >>> dreadArray out Fz
